//
//  AppDelegate.swift
//  RCServo
//
//  Created by Phidgets on 2018-02-27.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = RCServo()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var motorCalibrationBox: NSBox!
    @IBOutlet weak var minPulseWidthText: NSTextField!
    @IBOutlet weak var maxPulseWidthText: NSTextField!
    @IBOutlet weak var positionAtMinText: NSTextField!
    @IBOutlet weak var positionAtMaxText: NSTextField!
    @IBOutlet weak var motorDataBox: NSBox!
    @IBOutlet weak var motorPositionLabel: NSTextField!
    @IBOutlet weak var velocityEventBox: NSBox!
    @IBOutlet weak var motorVelocityLabel: NSTextField!
    @IBOutlet weak var stoppedCheck: NSButton!
    @IBOutlet weak var dataIntervalBox: NSBox!
    @IBOutlet weak var dataIntervalSlider: NSSlider!
    @IBOutlet weak var minDataIntervalLabel: NSTextField!
    @IBOutlet weak var maxDataIntervalLabel: NSTextField!
    @IBOutlet weak var dataIntervalLabel: NSTextField!
    @IBOutlet weak var motorControlsBox: NSBox!
    @IBOutlet weak var positionSlider: NSSlider!
    @IBOutlet weak var minTargetPositionLabel: NSTextField!
    @IBOutlet weak var maxTargetPositionLabel: NSTextField!
    @IBOutlet weak var targetPositionLabel: NSTextField!
    @IBOutlet weak var engagedButton: NSButton!
    @IBOutlet weak var advancedControls: NSBox!
    @IBOutlet weak var accelerationSlider: NSSlider!
    @IBOutlet weak var minAccelerationLabel: NSTextField!
    @IBOutlet weak var maxAccelerationLabel: NSTextField!
    @IBOutlet weak var accelerationLabel: NSTextField!
    @IBOutlet weak var velocityLimitSlider: NSSlider!
    @IBOutlet weak var minVelocityLimitLabel: NSTextField!
    @IBOutlet weak var maxVelocityLimitLabel: NSTextField!
    @IBOutlet weak var velocityLimitLabel: NSTextField!
    @IBOutlet weak var speedRampingCheck: NSButton!
    @IBOutlet weak var voltagePopup: NSPopUpButton!
    @IBOutlet weak var voltageBox: NSBox!
    
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.velocityChange.addHandler(velocitychange_handler)
            let _ = ch.positionChange.addHandler(positionchange_handler)
            let _ = ch.targetPositionReached.addHandler(targetreached_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.positionChange.removeAllHandlers()
            ch.velocityChange.removeAllHandlers()
            ch.targetPositionReached.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! RCServo
            let deviceID = try attachedDevice.getDeviceID()
            let engaged = try attachedDevice.getEngaged()
            let minPulseWidth = try attachedDevice.getMinPulseWidth()
            let maxPulseWidth = try attachedDevice.getMaxPulseWidth()
            var minDataInterval:UInt32?
            var maxDataInterval:UInt32?
            var dataInterval:UInt32?
                
    
            if(deviceID == DeviceID.PN_1061 || deviceID == DeviceID.PN_1066){
                minDataInterval = try attachedDevice.getMinDataInterval()
                maxDataInterval = try attachedDevice.getMaxDataInterval()
                dataInterval = try attachedDevice.getDataInterval()
            }
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.motorCalibrationBox.isHidden = false
                self.motorControlsBox.isHidden = false
                self.motorDataBox.isHidden = false
                self.dataIntervalBox.isHidden = true
                
                if(engaged){
                    self.engagedButton.title = "Disengage"
                }
                else{
                    self.engagedButton.title = "Engage"
                }
                
                self.minPulseWidthText.stringValue = String(format:"%.1f", minPulseWidth)
                self.maxPulseWidthText.stringValue = String(format:"%.1f", maxPulseWidth)
                self.stoppedCheck.isHidden = true
                
                switch(deviceID){
                case DeviceID.PN_1066:
                    fallthrough
                case DeviceID.PN_1061:
                    self.dataIntervalBox.isHidden = false
                    self.dataIntervalSlider.minValue = Double(minDataInterval!)
                    self.dataIntervalSlider.maxValue = Double(maxDataInterval!)
                    self.dataIntervalSlider.doubleValue = Double(dataInterval!)
                    self.minDataIntervalLabel.stringValue = String(format:"%d",minDataInterval!)
                    self.maxDataIntervalLabel.stringValue = String(format:"%d",maxDataInterval!)
                    self.dataIntervalLabel.stringValue = String(format:"%d",dataInterval!)
                    fallthrough
                case DeviceID.PN_RCC1000:
                    self.stoppedCheck.isHidden = false
                    self.velocityEventBox.isHidden = false
                    self.advancedControls.isHidden = false
                    if(deviceID == DeviceID.PN_RCC1000){
                        self.voltageBox.isHidden = false
                        self.voltagePopup.addItem(withTitle: "5V")
                        self.voltagePopup.addItem(withTitle: "6V")
                        self.voltagePopup.addItem(withTitle: "7.4V")
                    }
                    break
                default:
                    if(deviceID == DeviceID.PN_1001){
                        self.stoppedCheck.isHidden = false
                    }
                    self.advancedControls.isHidden = true
                    break
                }
                
                self.updateGUI(isInitial: true)
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                if(!self.advancedControls.isHidden){
                    frame.size.height += self.advancedControls.frame.size.height
                }
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.voltageBox.isHidden = true
            self.motorDataBox.isHidden = true
            self.motorControlsBox.isHidden = true
            self.advancedControls.isHidden = true
            self.motorCalibrationBox.isHidden = true
            
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func positionchange_handler(sender: RCServo, position: Double){
        DispatchQueue.main.async {
            self.motorPositionLabel.stringValue = String(format:"%.3f", position)
        }
    }
    
    func velocitychange_handler(sender: RCServo, velocity: Double){
        DispatchQueue.main.async {
            self.motorVelocityLabel.stringValue = String(format:"%.3f", velocity)
        }
    }
    
    func targetreached_handler(sender: RCServo, position: Double){
        DispatchQueue.main.async {
            self.stoppedCheck.state = NSControl.StateValue.on
            self.motorPositionLabel.stringValue = String(format:"%.3f", position)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func applyMotorCalibration(_ sender: Any) {
        do{
            let minPulseWidth = Double(minPulseWidthText.stringValue)
            let maxPulseWidth = Double(maxPulseWidthText.stringValue)
            let minPosition = Double(positionAtMinText.stringValue)
            let maxPosition = Double(positionAtMaxText.stringValue)
            if(minPulseWidth == nil || maxPulseWidth == nil || minPosition == nil || maxPosition == nil){
                outputError(errorDescription: "One or more of the calibration arguments is invalid.", errorCode: 0)
                return
            }
            try ch.setMinPulseWidth(minPulseWidth!)
            try ch.setMaxPulseWidth(maxPulseWidth!)
            try ch.setMinPosition(minPosition!)
            try ch.setMaxPosition(maxPosition!)
            self.updateGUI(isInitial: false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
        
    }
    
    @IBAction func setDataInterval(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setDataInterval(UInt32(slider.intValue))
            dataIntervalLabel.stringValue = String(format:"%d",slider.intValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setTargetPosition(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setTargetPosition(slider.doubleValue)
            stoppedCheck.state = NSControl.StateValue.off
            targetPositionLabel.stringValue = String(format:"%.3f",slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func engageMotor(_ sender: Any) {
        do{
            let engaged = try ch.getEngaged()
            try ch.setEngaged(!engaged)
            if(engaged){
                engagedButton.title = "Engage Motor"
            }
            else{
                engagedButton.title = "Disengage Motor"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setAcceleration(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setAcceleration(slider.doubleValue)
            accelerationLabel.stringValue = String(format:"%.3f",slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setVelocityLimit(_ sender: Any) {
        do{
            let slider = sender as! NSSlider
            try ch.setVelocityLimit(slider.doubleValue)
            velocityLimitLabel.stringValue = String(format:"%.3f",slider.doubleValue)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setSpeedRampingState(_ sender: Any) {
        do{
            let check = sender as! NSButton
            if(check.state == NSControl.StateValue.on){
                try ch.setSpeedRampingState(true)
                accelerationSlider.isEnabled = true
                velocityLimitSlider.isEnabled = true
            }
            else{
                try ch.setSpeedRampingState(false)
                accelerationSlider.isEnabled = false
                velocityLimitSlider.isEnabled = false
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setVoltage(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setVoltage(RCServoVoltage.volts_5_0)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setVoltage(RCServoVoltage.volts_6_0)
            }
            else{
                try ch.setVoltage(RCServoVoltage.volts_7_4)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
    
    //MARK: Misc
    func updateGUI(isInitial:Bool){
        do{
            let deviceID = try ch.getDeviceID();
            let minPosition = try ch.getMinPosition()
            let maxPosition = try ch.getMaxPosition()
            let minAcceleration = try ch.getMinAcceleration()
            let maxAcceleration = try ch.getMaxAcceleration()
            let acceleration = try ch.getAcceleration()
            let velocityLimit = try ch.getVelocityLimit()
            let minVelocityLimit = try ch.getMinVelocityLimit()
            let maxVelocityLimit = try ch.getMaxVelocityLimit()
            
            self.positionAtMinText.stringValue = String(format:"%.1f", minPosition)
            self.positionAtMaxText.stringValue = String(format:"%.1f", maxPosition)
            self.minTargetPositionLabel.stringValue = String(format:"%.1f", minPosition)
            self.maxTargetPositionLabel.stringValue = String(format:"%.1f", maxPosition)
            self.positionSlider.minValue = minPosition
            self.positionSlider.maxValue = maxPosition
            if(isInitial){
                motorPositionLabel.stringValue = "unknown"
            }
            else{
                let position = try ch.getPosition()
                motorPositionLabel.stringValue = String(format:"%.3f", position)
            }
            
            if(deviceID == DeviceID.PN_RCC1000 || deviceID == DeviceID.PN_1061 || deviceID == DeviceID.PN_1066){
                self.accelerationSlider.minValue = minAcceleration
                self.accelerationSlider.maxValue = maxAcceleration
                self.accelerationSlider.doubleValue = acceleration
                self.minAccelerationLabel.stringValue = String(format:"%.3f",minAcceleration)
                self.maxAccelerationLabel.stringValue = String(format:"%.3f",maxAcceleration)
                self.accelerationLabel.stringValue = String(format:"%.3f",acceleration)
                
                self.velocityLimitSlider.minValue = minVelocityLimit
                self.velocityLimitSlider.maxValue = maxVelocityLimit
                self.velocityLimitSlider.doubleValue = velocityLimit
                self.minVelocityLimitLabel.stringValue = String(format:"%.3f",minVelocityLimit)
                self.maxVelocityLimitLabel.stringValue = String(format:"%.3f",maxVelocityLimit)
                self.velocityLimitLabel.stringValue = String(format:"%.3f",velocityLimit)
            }
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
}

